# Made by Marcos Boggia

import cv2
from gui_automation.spot import Spot
import numpy as np
from pyautogui import screenshot
import imutils


TM_SQDIFF_NORMED = cv2.TM_SQDIFF_NORMED
TM_CCOEFF_NORMED = cv2.TM_CCOEFF_NORMED
TM_CCORR_NORMED = cv2.TM_CCORR_NORMED


def obtain_tm_results(img, tpl, method):
    result = cv2.matchTemplate(img, tpl, method)
    min_val, max_val, min_loc, max_loc = cv2.minMaxLoc(result)
    if method in [cv2.TM_SQDIFF, cv2.TM_SQDIFF_NORMED]:
        similarity = abs(min_val - 1)
        position = min_loc
    else:
        similarity = max_val
        position = max_loc
    return similarity, Spot(position, tpl.shape)


class Detection:

    def __init__(self, tpl, img=None):
        self.tpl = cv2.cvtColor(tpl, cv2.COLOR_BGR2GRAY)
        if img is None:
            self.img = cv2.cvtColor(np.array(screenshot())[:, :, ::-1].copy(), cv2.COLOR_BGR2GRAY)
        else:
            self.img = img

    def _apply_binary_thresh(self):
        _, img = cv2.threshold(self.img, 127, 255, cv2.THRESH_BINARY)
        _, tpl = cv2.threshold(self.tpl, 127, 255, cv2.THRESH_BINARY)
        return img, tpl

    # OpenCV Template Matching

    def tm(self, method=cv2.TM_SQDIFF_NORMED, thresh=False):
        if thresh:
            img, tpl = self._apply_binary_thresh()
        else:
            img, tpl = self.img, self.tpl
        return obtain_tm_results(img, tpl, method)

    def tm_multiscaled(self, method=cv2.TM_SQDIFF_NORMED, thresh=False, reduce_sc=0.2, magnify_sc=2.0, cant_sc=40):
        if thresh:
            img, tpl = self._apply_binary_thresh()
        else:
            img, tpl = self.img, self.tpl
        similarity, spot = obtain_tm_results(img, tpl, method)

        tpl_height, tpl_width = tpl.shape[:2]
        # Iterate through scales, resizing the image/screen and find better match for tpl.
        # Reducing up to reduce_sc. Defualt is 0.2, meaning it will resize image to a 0.2 of its size
        # Magnifying up to magnify_sc. Defualt is 2.0, meaning it will enlarge the image up to a double of its size
        for scale in np.linspace(reduce_sc, magnify_sc, cant_sc)[::-1]:
            # Resize
            resized = imutils.resize(img, width=int(img.shape[1] * scale))
            # if the resized image is smaller than the template, then break from the loop
            if resized.shape[0] < tpl_height or resized.shape[1] < tpl_width:
                break
            # apply tm
            new_similarity, new_spot = obtain_tm_results(resized, tpl, method)
            # keep the best match
            if new_similarity > similarity:
                similarity = new_similarity
                # Fix resized image coordinates.
                new_spot.revert_scaled_position_error(scale)
                spot = new_spot

        return similarity, spot



# Made by Marcos Boggia
# Official Doc: https://pyautogui.readthedocs.io/en/latest/mouse.html

import pyautogui
from time import sleep

pyautogui.PAUSE = 1
pyautogui.FAILSAFE = False


def click(x, y, clicks=1):
    pyautogui.click(x=x, y=y, button='left', clicks=clicks, interval=0.5)


def move(x, y):
    pyautogui.moveTo(x=x, y=y)


def hold_click(x, y, time):
    pyautogui.PAUSE = 0
    refresh_rate = 0.05
    pyautogui.mouseDown(x=x, y=y, button='left')
    while time > 0:
        sleep(refresh_rate)
        pyautogui.moveTo(x=x, y=y)
        time -= refresh_rate

    pyautogui.mouseUp(x=x, y=y, button='left')
    pyautogui.PAUSE = 1


def drag_click(start_x, start_y, end_x, end_y):
    pyautogui.moveTo(x=start_x, y=start_y)
    pyautogui.dragTo(x=end_x, y=end_y, button='left', duration=1)



from gui_automation.gui_automation import GuiAuto
from gui_automation.image_detection import TM_SQDIFF_NORMED, TM_CCOEFF_NORMED, TM_CCORR_NORMED
from gui_automation.image_loader import load_images
name = "gui_automation"



# Made by Marcos Boggia


# Wraps all position/coordinates calculations for the found spot.
class Spot:
    """
    Spot class, wraps the position of a the found image.

    Available methods:
    upper_left_position()
    upper_right_position()
    bottom_left_position()
    bottom_right_position()
    center_position()
    custom_position(x_multiplier, x_modifier, y_multiplier, y_modifier)

    """
    def __init__(self, position, tpl_shape):
        self.position = position
        self.tpl_shape = tpl_shape

    # Position calculations

    def upper_left_position(self):
        return self.position

    def upper_right_position(self):
        new_y = self.position[1] + self.tpl_shape[0]
        return self.position[0], new_y

    def bottom_left_position(self):
        new_x = int(self.position[0] + self.tpl_shape[1])
        return new_x, self.position[1]

    def bottom_right_position(self):
        new_x = self.position[0] + self.tpl_shape[1]
        new_y = self.position[1] + self.tpl_shape[0]
        return new_x, new_y

    def center_position(self):
        new_x = int(self.position[0] + self.tpl_shape[1] / 2)
        new_y = int(self.position[1] + self.tpl_shape[0] / 2)
        return new_x, new_y

    def custom_position(self, x_multiplier, x_modifier, y_multiplier, y_modifier):
        """
        This function helps calculate any coordinate inside the image to be detected.
        Ej: x, y = custom_position(3, 8, 1, 2)
           3/8 of the width
          __o_____
         |        |
         |        |
         |  x     o 1/2 of the height
         |        |
         |________|

        :param x_multiplier: how many parts of the divided width to take
        :param x_modifier: in how many parts the width is going to be divided
        :param y_multiplier: same as width but with height
        :param y_modifier: same as width but with height
        :return:
        new_x, new_y: the calculated coordinates
        """
        new_x = int(self.position[0] + self.tpl_shape[1] / x_modifier * x_multiplier)
        new_y = int(self.position[1] + self.tpl_shape[0] / y_modifier * y_multiplier)
        return new_x, new_y

    def revert_scaled_position_error(self, scale):
        """
        If image has been resized this function must be used to fix the displacement of the position to the original
        image.
        :param scale: the scale used to resize the image
        :return: No return
        """
        x = self.position[0]
        y = self.position[1]
        self.position = (int(x/scale), int(y/scale))



# Made by Marcos Boggia

import gui_automation.image_detection as imgd
import gui_automation.mouse as mouse
import functools


def _detect(func):
    """

    :param func: behaviour to inject if the detection fits the similarity threshold
    :return: decorated function
    """
    @functools.wraps(func)
    def wrapper(self, *args, **kwargs):
        det = imgd.Detection(self.tpl)
        if self.multiscaled:
            if self.multiscaled is True:
                self.similarity, self.spot = det.tm_multiscaled(self.method, self.thresh)
            else:
                self.similarity, self.spot = det.tm_multiscaled(self.method, self.thresh, *self.multiscaled)
        else:
            self.similarity, self.spot = det.tm(self.method, self.thresh)
        if self.similarity > self.similarity_threshold:
            func(self, *args, **kwargs)
            return True
        self.spot = None
        return False

    return wrapper


class GuiAuto:
    """
    Detects and image withing the screen and performs an action.

    Methods
    ----------
    update: used to replace image to be found, similarity threshold and other parameters.
    detect: returns True if it finds the tpl in the image.
    detect_and_move: same as detect but it moves the cursor to the center of the found image withing the screen.
    detect_and_click: Clicks the left buttons the quantity specified in @param click(default 1) in the center of the
                      found image.
    detect_and_hold: same as detect_and_click but instead of clicking X times, it holds the click @param time seconds.
    detect_and_drag: Drags from one point to another. For more information read this function docstring.

    """

    def __init__(self, tpl=None, similarity_threshold=None, method=imgd.TM_CCOEFF_NORMED, thresh=False, multiscaled=False):
        """
        Parameters
        ----------
        tpl : image/numpy matrix with pixels.
            The image to be found.
        similarity_threshold:
            It goes from 0 (no match at all) to 1 (perfect match).

        Optional parameters
        ----------
        method: OpenCV TM methods: TM_SQDIFF_NORMED, TM_CCOEFF_NORMED, TM_CCORR_NORMED
            OpenCV template match method to use. Default is TM_SQDIFF_NORMED.
        thresh:
            Apply a binary threshold to images before detection. Default is False.
        multiscaled:
            Applies template match multiple times with different scales of the screen.
        """
        self.tpl = tpl
        self.similarity_threshold = similarity_threshold
        self.method = method
        self.thresh = thresh
        self.multiscaled = multiscaled
        self.similarity = None
        self.spot = None

    def update(self, tpl, similarity_threshold, method=imgd.TM_CCOEFF_NORMED, thresh=False, multiscaled=False):
        """
        Same parameters as class instantiation. Used to update image to be found and the similarity threshold (instead
        of creating a new instance with these new values) and other parameters optionally.
        """
        self.tpl = tpl
        self.similarity_threshold = similarity_threshold
        self.method = method
        self.thresh = thresh
        self.multiscaled = multiscaled
        return self

    @_detect
    def detect(self):
        pass

    @_detect
    def detect_and_move(self):
        mouse.move(*self.spot.center_position())

    @_detect
    def detect_and_click(self, clicks=1):
        mouse.click(*self.spot.center_position(), clicks)

    @_detect
    def detect_and_hold(self, time):
        mouse.hold_click(*self.spot.center_position(), time)

    @_detect
    def detect_and_drag(self, start_x_fraction, start_y_fraction, end_x_fraction, end_y_fraction):
        """
        @brief Drags the mouse from one point to another using the tpl width and height to calculate starting and ending
                points. All params are fractions in the following string format: 'number/number'.
        @example GuiAuto(img, 0.8).detect_and_drag('3/4', '0/1', '7/8', '4/5')

           3/4 of the width and 0/1 of the height for START
          __o___o_  7/8 of the width for END
         |  S     |   S = start
         |   \\    |   E = end
         |    \\   |   \\ = the mouse drag path
         |      E o 4/5 of the height for END
         |________|
        """
        start_x, start_y = self.spot.custom_position(*_values_from_fraction(start_x_fraction),
                                                     *_values_from_fraction(start_y_fraction))
        end_x, end_y = self.spot.custom_position(*_values_from_fraction(end_x_fraction),
                                                 *_values_from_fraction(end_y_fraction))
        mouse.drag_click(start_x, start_y, end_x, end_y)


# Used to obtain the numbers from fractions like '3/4' etc.
def _values_from_fraction(fraction):
    return int(fraction[0]), int(fraction[2])
